/* usaf - Ulam Spiral (animation) Fun */

#version 3.8;

global_settings {assumed_gamma 1}

/* TODO clock-based spline? */
camera {
  location <0,6,-17.5>
  right x * (16/10)
  up y
  angle 60
  look_at <2,.75,2>
}


#declare verbose_ = false;
//#declare verbose_ = true;

//#declare log_frames_ = false;
#declare log_frames_ = true;


light_source {<-1,1,-1> * 1e3 color srgb 1 parallel}
light_source {<1,1,-1> * 1e3 color srgb 1 parallel}

/* the Wikipedia article suggests that '1' on the spiral can, in fact,
 * be any integer, they give example '41'; requires corresponding csv
 * and setting start/end frames in ini.
 */
#declare addend_ = 0;


#include "filed.inc"
#include "foreach.inc"
#include "ruled.inc"
#include "usaf.inc"

/* file/record layouts.
 * items:  {type, point}
 * logger: {frame, {state}}
 * state:  {prev, curr, face, stride, leg, n}
 */
#declare itmrcd_ = array [2] {"I","V33"};
#declare logrcd_ = array [7] {"I","V20","V20","I","I","I","I"};
#declare objrcd_ = array [6] {"V20","V20","I","I","I","I"};

/* small primes lookup, load before setting working directory.
 * set "is prime" flag/index.
 */
#declare primes_ = dictionary {
  .File     : "usaf.csv",
  .Access   : "r",
  .Fields   : array [1] {"I"},
  .hasNames : true
};
Filed(primes_)
/* start with frame 0, hence ... */
#declare ix_ =
    (usaf__isPrime(primes_.Data,(1 + frame_number + addend_)) ? 1 : 0);

/* ideally use RAM disk.  want same path in .ini file */
#declare fild_workingDir = "/tmp/render/";

#if (frame_number)
  // check have all files.
  #if (!usaf__chkFilesExist(true))
    #error "oops, missing state and or items files."
  #end
  // retrieve state.
  #declare state_ = dictionary {
    .File    : usaf__data_files_[0],
    .Access  : "r",
    .Fields  : objrcd_,
    .Verbose : verbose_
  };
  Filed(state_)
  // step object.
  usaf__obj(state_.Data[0])
  // store new state.
  #local state_.Access = "w";
  #local state_.Strict = false;
  Filed(state_)
  // append items.
  #declare items_ = dictionary {
    .File    : usaf__data_files_[1],
    .Access  : "a",
    .Fields  : itmrcd_,
    .Data    : array
        {array mixed [2] {ix_, (usaf__pos2point(state_.Data[0][1]))}},
    .Verbose : verbose_
  };
  Filed(items_)
  // append log.
  #if (log_frames_)
    #declare logger_ = dictionary {
      .File     : usaf__data_files_[2],
      .Access   : "a",
      .Fields   : logrcd_,
      .Data     : array,
      .Verbose  : verbose_
    };
    usaf__add2log(logger_,frame_number,state_.Data[0])
    Filed(logger_)
  #end
#else
  // frame #0 setup, check no files yet.
  #if (!usaf__chkFilesExist(false))
    #error "oops, found stale state and or items files."
  #end
  // create state, items, and log if used.
  #declare items_ = dictionary {
    .File    : usaf__data_files_[1],
    .Access  : "w",
    .Fields  : itmrcd_,
    .Data    : array {array mixed [2] {0, (usaf__pos2point(<0,0>))}},
    .Verbose : verbose_
  };
  Filed(items_)
  // nb. algorithm relies on exactly those initial values.
  #declare state_ = dictionary {
    .File    : usaf__data_files_[0],
    .Access  : "w",
    .Fields  : objrcd_,
    .Data    : array {array mixed [6] {<0,0>,<0,0>,1,1,1,1}},
    .Verbose : verbose_
  };
  Filed(state_)
  #if (log_frames_)
    #declare logger_ = dictionary {
      .File     : usaf__data_files_[2],
      .Access   : "w",
      .Fields   : logrcd_,
      .hasNames : true,
      .Names    : array [7] {"frm#","prev","curr","face","stride","leg","n"},
      .Data     : array,
      .Verbose  : verbose_
    };
    usaf__add2log(logger_,frame_number,state_.Data[0])
    Filed(logger_)
  #end
#end

Ruled(dictionary {
  .norm : <0,1,0>,
  .offs : -.05,
  .divs : 1,
  .bg   : pigment {color rgbt .975},
  .dim  : false
})

/* read up-to-date items.  make line begin in "floor" */
#declare items_ = dictionary {
  .File    : usaf__data_files_[1],
  .Access  : "r",
  .Fields  : itmrcd_,
  .Verbose : verbose_
};
Filed(items_)

#declare tmp_ = dictionary {.last_: <.5,-.1,.5>};

Foreach(items_.Data, dictionary {.Macro: "usaf__mkItems", .Arg: "tmp_"})

object {
  usaf__data_obj_
  pigment {color rgb <1,0,0>}
  rotate <0,usaf__data_rotates_[state_.Data[0][2]],0>
  translate <(state_.Data[0][1].x+.5),0,(state_.Data[0][1].y+.5)>
}

